from time import sleep, time
import re
#from clients.client import Client
import logging, coloredlogs
from sqlalchemy import Column, Integer, String, Float, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine
from threading import Lock, Thread

import paho.mqtt.client as mqtt
import socket
import struct
import os

db_lock = Lock()

# Middle: serial-A502C1P4
#NODES = ['serial-A502C1OZ', 'serial-A502C1PH', 'serial-A502C1PV', 'serial-A502C1PE', 'serial-A502C1Q1', 'serial-A502C1QH', 'serial-A502C1P5', 'serial-A502C1PI', 'serial-A502C1P4']
# node ids: 7, ..., 16
NODES = ['serial-A502C1PO', 'serial-A502C1PV', 'serial-A502C1P4', 'serial-A502C1QH', 'serial-A502C1OZ', 'serial-A502C1PE', 'serial-A502C1PI', 'serial-A502C1P5', 'serial-A502C1PH', 'serial-A502C1Q1']
CHANNELS = [11, 15, 26]
ANGELS = [0,25,50,75,100,125,150,175]
POWERS = [-17, -7, 3]
PAYLOADS = ['short', 'looooooooooooong']

LOG_DIRECTORY = "/data/experiment-temperature/"

if False:
    CHANNELS = [11]
    POWERS = [-17]
    PAYLOADS = ['short']

#logger = Logger(__file__)
Base = declarative_base()

class Round(Base):
    __tablename__ = 'round'
    id = Column(Integer, nullable=False, primary_key=True, autoincrement=True)
    channel = Column(Integer, nullable=False)
    power = Column(Integer, nullable=False)
    angle = Column(Integer, nullable=False)
    payload = Column(String(250), nullable=False)
    ts = Column(Float, nullable=False)

class TX(Base):
    __tablename__ = 'tx'
    serial_id = Column(String(250), nullable=False, primary_key=True)
    ts = Column(Float, nullable=False, primary_key=True)
    seq_nr = Column(Integer, nullable=False)
    round_id = Column(Integer)

class RX(Base):
    __tablename__ = 'rx'
    serial_id = Column(String(250), nullable=False, primary_key=True)
    ts = Column(Float, nullable=False, primary_key=True)
    rssi = Column(Integer, nullable=False)
    lqi = Column(Integer, nullable=False)
    node_id = Column(Integer, nullable=False)
    seq_nr = Column(Integer, nullable=False)
    round_id = Column(Integer)

class Corrupt(Base):
    __tablename__ = 'corrupt'
    serial_id = Column(String(250), nullable=False, primary_key=True)
    ts = Column(Float, nullable=False, primary_key=True)
    rssi = Column(Integer, nullable=False)
    lqi = Column(Integer, nullable=False)
    payload = Column(String(255), nullable=False)
    round_id = Column(Integer)

engine = create_engine('sqlite:////data/row_2.db')
Session = sessionmaker(bind=engine)
#Base.metadata.create_all(engine)

re_tx = re.compile("^>(?P<seq_nr>[0-9]+)=(?P<checksum>[A-Fa-f0-9]+)$")
re_rx = re.compile("^<(?P<node_id>[0-9]+)\|(?P<seq_nr>[0-9]+)\|(?P<rssi>\-?[0-9]+)\|(?P<lqi>[0-9]+)=(?P<checksum>[A-Fa-f0-9]+)$")
# <-94|52|0|=A5
re_rx_corrupt = re.compile("^<(?P<rssi>-?[0-9]+)\|(?P<lqi>[0-9]+)\|(?P<length>[0-9]+)\|(?P<payload>(?: [A-Fa-f0-9]+)*)=(?P<checksum>[A-Fa-f0-9]+)$")

re_ifconfig = re.compile("^success: set (.*?) o(n|f) interface [0-9]+ to (.*?)$")

eval_round = None

def calculate_checksum(*args):
    s = 0
    for a in args:
        s += sum(bytes(a, 'utf-8'))
    return hex( s&0xFF )[2:].upper().zfill(2)

def handle_match_tx( src_id, ts, match_tx ):
    seq_nr = match_tx.group("seq_nr")
    checksum = match_tx.group("checksum")
    
    expected_checksum = calculate_checksum( seq_nr )
    if expected_checksum == checksum:
        db_lock.acquire()
        session = Session()
        session.add(TX( round_id=eval_round, serial_id = src_id, ts=ts, seq_nr = seq_nr ))
        session.commit()
        session.close()
        db_lock.release()
        #logger.log("[{}] TX: #{}".format( src_id, seq_nr ))
        return True
    else:
        logger.debug("Corrupt Line")
        return False

def handle_match_rx( src_id, ts, match_rx ):
    node_id = match_rx.group("node_id")
    seq_nr = match_rx.group("seq_nr")
    rssi = match_rx.group('rssi')
    lqi = match_rx.group('lqi')
    checksum = match_rx.group("checksum")
    expected_checksum = calculate_checksum( node_id, '|', seq_nr, '|', rssi, '|', lqi )

    if expected_checksum == checksum:
        db_lock.acquire()
        session = Session()
        session.add(RX(round_id=eval_round, serial_id = src_id, ts=ts, seq_nr = seq_nr, node_id = node_id, rssi = rssi, lqi = lqi))
        session.commit()
        session.close()
        db_lock.release()
        # TODO(rh): Add sqlite3.OperationalError
        # TODO(rh): Add sqlalchemy.exc.OperationalError / sqlite3.OperationalError
        #logger.log("[{}] RX: #{} from {}".format( src_id, node_id, seq_nr ))
        return True
    else:
        logger.debug("Corrupt Line {} != {}".format( expected_checksum, checksum) )
        return False

def handle_match_rx_corrupt( src_id, ts, match_rx_corrupt ):
    rssi = match_rx_corrupt.group('rssi')
    lqi = match_rx_corrupt.group('lqi')
    payload = match_rx_corrupt.group('payload').strip()
    logger.debug( "Corrupt: {}dBm, {}, {}".format( rssi, lqi, payload ) )
    db_lock.acquire()
    session = Session()
    session.add(Corrupt(round_id=eval_round, serial_id = src_id, ts=ts, rssi = rssi, lqi = lqi, payload = payload))
    session.commit()
    session.close()
    db_lock.release()
    return True

def handle_serial_rx(d):
    global eval_round
    src_id = d['src']['id'][0]
    ts = d['ts']
    line = d['line'].strip()

    if eval_round is None:
        return False

    match_ifconfig = re_ifconfig.match(line)
    if match_ifconfig is not None:
        return True
    match_tx = re_tx.match(line)
    if match_tx is not None:
        return handle_match_tx( src_id, ts, match_tx )
    match_rx = re_rx.match(line)
    if match_rx is not None:
        return handle_match_rx( src_id, ts, match_rx )
    match_rx_corrupt = re_rx_corrupt.match(line)
    if match_rx_corrupt is not None:
        return handle_match_rx_corrupt( src_id, ts, match_rx_corrupt  )
    return False


class PRREvalClient(Thread):
    def __init__(self, args = None):
        super(PRREvalClient, self).__init__(target=self)
        #self.session = Session()
        self.logger = logging.getLogger( self.__class__.__name__)

        self.my_nodes = []

        self.connected = 0

        self.log_mutex = Lock()

        self.server_one = "pi-a279"
        self.HUB_ONE_DISCOVERY_TOPIC = "hotplug/{}/discovery".format(self.server_one)
        self.HUB_ONE_DISCOVERY_ANSWER_TOPIC = "hotplug/{}/discovery/answer".format(self.server_one)

        self.server_two = "pi-b827"
        self.HUB_TWO_DISCOVERY_TOPIC = "hotplug/{}/discovery".format(self.server_two)
        self.HUB_TWO_DISCOVERY_ANSWER_TOPIC = "hotplug/{}/discovery/answer".format(self.server_two)

        self.client_one = mqtt.Client()
        self.client_one.connect( "pi-a279.dyn", port=1883, keepalive=5 )
        self.client_one.user_data_set("pi-a279") #use userdata to keep the clients apart!
        self.client_one.on_connect = self.on_connect
        self.client_one.on_message = self.on_message
        self.client_one.on_disconnect = self.on_disconnect
        self.client_one.loop_start()

        self.client_two = mqtt.Client()
        self.client_two.connect( "pi-b827.dyn", port=1883, keepalive=5 )
        self.client_two.user_data_set("pi-b827") #use userdata to keep the clients apart!
        self.client_two.on_connect = self.on_connect
        self.client_two.on_message = self.on_message
        self.client_two.on_disconnect = self.on_disconnect
        self.client_two.loop_start()

        self.interface = 3

        self.file_formatter = logging.Formatter('%(message)s')
        self.logger.debug("Init done")


    def on_disconnect(self, client, userdata, rc):
        print(" [on_disconnect] {}".format( rc ))
        self.connected -= 1

    def on_connect(self, client, userdata, flags, rc):
        if userdata == "pi-a279":
            self.client_one.subscribe(self.HUB_ONE_DISCOVERY_ANSWER_TOPIC, qos=2)
            self.logger.debug("Subscribed to Topic: {}".format(self.HUB_ONE_DISCOVERY_ANSWER_TOPIC))
            self.connected += 1
        elif userdata == "pi-b827":
            self.client_two.subscribe(self.HUB_TWO_DISCOVERY_ANSWER_TOPIC, qos=2)
            self.logger.debug("Subscribed to Topic: {}".format(self.HUB_TWO_DISCOVERY_ANSWER_TOPIC))
            self.connected += 1
        else:
            self.logger.warning("Connected to unexspected mqtt server!")
        

    def on_message(self, client, userdata, msg):
        topic = msg.topic.split('/')
        if msg.topic == self.HUB_ONE_DISCOVERY_ANSWER_TOPIC or msg.topic == self.HUB_TWO_DISCOVERY_ANSWER_TOPIC:
            all_nodes = str(msg.payload)[2:][:-1]
            all_nodes = all_nodes.split(";")
            self.logger.debug("Got nodes: {}".format(all_nodes))
            for node in all_nodes:
                if node not in self.my_nodes and node is not '':
                    rx_topic = ""
                    if msg.topic == self.HUB_ONE_DISCOVERY_ANSWER_TOPIC:
                        self.my_nodes.append( (node, self.server_one) )
                        rx_topic = "hotplug/{}/{}/rx".format(self.server_one, node)
                    else:
                        self.my_nodes.append( (node, self.server_two) )
                        rx_topic = "hotplug/{}/{}/rx".format(self.server_two, node)
                    #subscribe on client that started callback
                    client.subscribe(rx_topic, qos=2)
                    self.logger.debug("Subscribing to {}".format(rx_topic))
            print("My Nodes: {}".format(self.my_nodes))

        elif topic[0] == "hotplug" and len(topic) == 4:
            #create dir for Logs files if it does not exist yet!
            if not os.path.isdir("{}/{}".format(LOG_DIRECTORY, userdata)):
                os.mkdir("{}/{}".format(LOG_DIRECTORY, userdata))
            unpacked_msg = str(msg.payload)[2:][:-1]
            node_id = topic[2]

            self.log_mutex.acquire()
            
            fh = logging.FileHandler("{}/{}/{}.log".format(LOG_DIRECTORY, userdata, node_id))
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(self.file_formatter)

            self.logger.addHandler(fh)
            self.logger.info("{:0<20},{},{}".format(time(), node_id, unpacked_msg))
            self.logger.removeHandler(fh)

            self.log_mutex.release()
        else:
            print( "[on_message] {} {}".format(msg.payload, msg.topic) )

    def eval(self):
        #wait until both clients are connected!
        while self.connected < 2:
            pass
        
        #if we have no nodes, subscribe to discovery topics
        if len(self.my_nodes) != 14:
            self.logger.debug("Publishing message on topic: {}".format(self.HUB_ONE_DISCOVERY_TOPIC))
            self.client_one.publish(self.HUB_ONE_DISCOVERY_TOPIC, struct.pack("d", 5), qos=2)

            self.logger.debug("Publishing message on topic: {}".format(self.HUB_TWO_DISCOVERY_TOPIC))
            self.client_two.publish(self.HUB_TWO_DISCOVERY_TOPIC, struct.pack("d", 5), qos=2)

            #wait until we have 14 Nodes (7 in each chamber)
            self.logger.info("Waiting for Node Data!")
            while len(self.my_nodes) != 14:
                pass
            self.logger.info("Got Data of all Nodes!")

        global eval_round
        self.logger.info("Hallo")
        for channel in CHANNELS:
            self.logger.info( "CHANNEL={} ".format(channel) )
            self.set_channel( channel  )
            sleep( 5 )
            for power in POWERS:
                self.logger.info( "POWER={} ".format(power) )
                self.set_txpower( power  )
                sleep( 1 )
                for payload in PAYLOADS:
                    self.logger.info( "PAYLOAD={} ".format(payload) )
                    self.send_all( "payload {}".format( payload )  )
                    sleep( 5 )
                    db_lock.acquire()
                    r = Round( angle=-1, channel=channel, power=power, payload=payload, ts=time() )
                    #self.session.add( r )
                    #self.session.commit()
                    eval_round = r.id
                    db_lock.release()
                    for sender_info in self.my_nodes:
                        #self.send({'dst': {'id': sender}, 'type': 'serial_tx', 'line': 'tx 5 200000\n'})
                        self.send(sender_info, "tx 5 2000")
                        # 5 packets, 200ms -> 1 second
                        sleep( 5 )
                        #self.send({'dst': {'id': sender}, 'type': 'serial_tx', 'line': 'ifconfig 3 set state idle\n'})
                        #self.send(sender, 'ifconfig 3 set state idle')
                        self.send(sender_info, 'ifconfig {} set state idle'.format(self.interface))
                        sleep( 5 )
                    # Delay after all sends
                    sleep( 5.0 )

    def run(self):
        self.running = True
        while self.running:
            try:
                self.eval()
                
                fh = logging.FileHandler("{}/master-log.log".format(LOG_DIRECTORY))
                fh.setLevel(logging.DEBUG)
                fh.setFormatter(self.file_formatter)

                self.log_mutex.acquire()

                self.logger.addHandler(fh)
                self.logger.info("\n")
                self.logger.info("\n")
                self.logger.removeHandler(fh)

                self.log_mutex.release()

                break

            except KeyboardInterrupt:
                self.logger.debug("Stopped")
                self.running = False
                break

    def stop(self):
        if self.running:
            self.logger.info("stop()")
            self.running = False
            return True
        else:
            self.logger.warn("Unable to stop: not alive")
            return False

    def set_channel(self, channel):
        #self.send_all("ifconfig 3 set state idle\n")
        #self.send_all("ifconfig 3 set channel {}".format( channel ))
        self.send_all("ifconfig {} set channel {}".format( self.interface, channel ))

    def set_txpower(self, power):
        #self.send_all("ifconfig 3 set power {}".format( power ))
        self.send_all("ifconfig {} set power {}".format( self.interface, power ))

    def send_all(self, line):
        for sender in self.my_nodes:
            self.send(sender, line)
        #self.send({'dst': {'id': self.my_nodes}, 'type': 'serial_tx', 'line': line})

    def send(self, sender_info, line):
        (node_id, server_name) = sender_info
        tx_topic = "hotplug/{}/{}/tx".format(server_name, node_id)
        self.logger.warning("Sending to node {}: {}".format(node_id, line))
        #self.logger.warning("{}".format(line))
        if server_name == self.server_one:
            self.client_one.publish(tx_topic, bytes(line, 'utf-8'), qos=2)
        else:
            self.client_two.publish(tx_topic, bytes(line, 'utf-8'), qos=2)

        fh = logging.FileHandler("{}/master-log.log".format(LOG_DIRECTORY))
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(self.file_formatter)

        self.log_mutex.acquire()

        self.logger.addHandler(fh)
        self.logger.info("{:0<20},{:>30},{}".format(time(), node_id, line))
        self.logger.removeHandler(fh)

        self.log_mutex.release()
        pass

if __name__ == '__main__':
    #create directory for log files!
    if not os.path.isdir(LOG_DIRECTORY):
        os.mkdir(LOG_DIRECTORY)
    coloredlogs.install(fmt='%(asctime)s %(name)s %(levelname)s %(message)s', level=0)
    client = PRREvalClient()
    client.run()
