

#!/usr/bin/env python3

import paho.mqtt.client as paho
import paho.mqtt.publish as publish
from time import sleep, time
import socket
import struct
import configparser
from threading import Event

import os

CHAMBER_ONE_NAME = "lctc-c01"
CHAMBER_TWO_NAME = "lctc-c04"

client_one = paho.Client(client_id="temp_logger_" + CHAMBER_ONE_NAME, clean_session=True)
client_two = paho.Client(client_id="temp_logger_" + CHAMBER_TWO_NAME, clean_session=True)

CHAMBER_ONE_TEMPERATURE_TOPIC = "ds1820/lctc-c01/28-000007fb4c59/temp"
CHAMBER_TWO_TEMPERATURE_TOPIC = "ds1820/lctc-c04/28-000007fbfbcb/temp"

LOG_DIRECTORY = "/data/experiment-temperature/"

CHAMBER_ONE_LOG_FILE = LOG_DIRECTORY + "lctc-c01-temp-log.csv"
CHAMBER_TWO_LOG_FILE = LOG_DIRECTORY + "lctc-c04-temp-log.csv"

def on_message(client, userdata, msg):
    if msg.topic == CHAMBER_ONE_TEMPERATURE_TOPIC:
        temp, = struct.unpack("d", msg.payload)
        with open(CHAMBER_ONE_LOG_FILE, 'a') as fh:
            fh.write("{},{}\n".format(time(), temp))

    elif msg.topic == CHAMBER_TWO_TEMPERATURE_TOPIC:
        temp, = struct.unpack("d", msg.payload)
        with open(CHAMBER_TWO_LOG_FILE, 'a') as fh:
            fh.write("{},{}\n".format(time(), temp))
        
    else:
        print("[on_message] {} {}".format(msg.payload, msg.topic) )

def on_disconnect(client, userdata, rc):
    print("[on_disconnect] {}".format( rc ))

def on_connect(client, userdata, flags, rc):
    if userdata == CHAMBER_ONE_NAME:
        client.subscribe(CHAMBER_ONE_TEMPERATURE_TOPIC, qos=2)
        print("Subscribed to topic: {}".format(CHAMBER_ONE_TEMPERATURE_TOPIC))
    elif userdata == CHAMBER_TWO_NAME:
        client.subscribe(CHAMBER_TWO_TEMPERATURE_TOPIC, qos=2)
        print("Subscribed to topic: {}".format(CHAMBER_TWO_TEMPERATURE_TOPIC))
    else:
        print("Connected to unknwon broker!")
    

def set_temp(temp):
    client.publish("chamber/"+CHAMBER_NAME+"/temp", struct.pack("d", temp), qos=2)
    temperature_reached.clear() #Set Event flag to false and Block until set is called

if __name__ == "__main__":
    #if not os.path.isdir(CHAMBER_ONE_LOG_FILE):
    #    os.mkdir(CHAMBER_ONE_LOG_FILE)
		
    SERVERNAME = "{}.ibr.cs.tu-bs.de".format(CHAMBER_ONE_NAME)
    print(SERVERNAME)
    client_one.connect( SERVERNAME, port=1883, keepalive=5 )
    client_one.user_data_set(CHAMBER_ONE_NAME)
    client_one.on_connect = on_connect
    client_one.on_message = on_message
    client_one.on_disconnect = on_disconnect
    client_one.loop_start()

    SERVERNAME = "{}.ibr.cs.tu-bs.de".format(CHAMBER_TWO_NAME)
    print(SERVERNAME)
    client_two.connect( SERVERNAME, port=1883, keepalive=5 )
    client_two.user_data_set(CHAMBER_TWO_NAME)
    client_two.on_connect = on_connect
    client_two.on_message = on_message
    client_two.on_disconnect = on_disconnect
    client_two.loop_start()

    while True:
        pass
